"""
Solar Position and Irradiance Calculations

This module provides functions for calculating solar position parameters and clear-sky solar irradiance on a horizontal surface. The implemented equations provide approximations without accounting for atmospheric effects and more complex Earth--Sun patterns. It aims to facilitate computing power and energy generated by PV modules at any given location and time.

Introduction to scientific programming FHV--NES 2023
"""

# Importing packages at the top of your script is not required, but considered
# good practice. The order doesn't matter, but here again the python police
# recommends to use the following order:
#   - standard library packages
#   - installed packages
#   - library modules
import numpy as np
import pandas as pd

LAT_DO = 47.4124
LON_DO = 9.7438

def comp_declination(day_of_year):
    """
    Calculate the solar declination.

    Parameters
    -----------
    day_of_year: array-like, int
        Day of the year (1 to 365).

    Returns
    --------
    out : array-like, float
        Solar declination in degrees.

    Examples
    --------
    >>> comp_declination(80)
    0.0
    >>> comp_declination(265)
    -1.0088713639562366
    >>> comp_declination(172)
    23.448045645453604
    >>> comp_declination(355)
    -23.44457137142844
    """
    epsilon = 23.45
    declination = epsilon * np.sin(2 * np.pi * (day_of_year - 80) / 365)
    return declination


def convert_utc2localsolar(utc_series, lon):
    """
    Convert UTC time to local solar time based on longitude. 
    
    The conversion does not take into account the equation of time (i.e., day of year, etc),
    so it only represents a simple approximation.
    
    Parameters
    -----------
    utc_series: pd.Series
        Pandas Series of dtype datetime containing UTC
    lon: float

    Returns
    -------
    local_series: pd.Series
        Pandas DatetimeIndex, of the local solar datetime.
    """
    time_difference = pd.to_timedelta(lon/15, unit='h')
    local_series = pd.to_datetime(utc_series) + time_difference

    return local_series


def comp_hourangle(time_series, tzone="localsolar", lon=None):
    """
    Convert time to hour angle. 
    
    Parameters
    -----------
    time_series: pd.Series
        Pandas Series of dtype datetime containing UTC
    tzone: {'utc', 'localsolar'}, optional
        Does the time_series represent UTC time or local solar time?
    lon: array-like, float
        longitude (required if tzone='utc')

    Returns
    -------
    tau: array-like, float
        Array of hour angle (negative in the morning, 
        positive in the afternoon, 0 at noon)
    """
    if tzone == "utc":
        time = convert_UTC2local(time_series, lon)
    elif tzone == "localsolar":
        time = time_series
    else:
        raise ValueError("tzone can only be 'localsolar' or 'utc'")
        
    # tau is shifted in relation to time by half a day:
    time_shifted = time - pd.to_timedelta(12, unit='h')
    # convert time to angle (involves converting the hour and minute to a float)
    tau = pd.Series(15*(time_shifted.hour + time_shifted.minute/60))
    # shift tau from morning hours to ensure
    # negative tau in the morning, positive in the afternoon, 0 at noon
    tau[tau >= 180] = tau[tau >= 180] - 360

    return tau.values


def comp_elevangle(hourangle, declination, lat):
    """
    Compute the elevation angle of the sun in degrees

    Parameters
    ----------
    hourangle: array-like, float
        hour angle tau in degrees (negative in the morning, 
        positive in the afternoon, 0 at noon)
    declination: array-like, float
        in degrees
    lat: array-like, float
        latitude in degrees

    Returns
    -------
    h: array-like, float
        elevation angle in degrees
    """
    h = np.rad2deg(np.arcsin(
        np.cos(np.deg2rad(declination)) * np.cos(np.deg2rad(hourangle)) * np.cos(np.deg2rad(lat)) + 
        np.sin(np.deg2rad(declination)) * np.sin(np.deg2rad(lat))
    ))
    return h


def comp_solarazimuth(elevangle, hourangle, declination, lat):
    """
    Compute the azimuth of the sun in degrees

    Parameters
    ----------
    elevangle: array-like, float
        elevation angle of the sun in degrees
    hourangle: array-like, float
        hour angle tau in degrees (negative in the morning, 
        positive in the afternoon, 0 at noon)
    declination: array-like, float
        in degrees
    lat: array-like, float
        latitude in degrees

    Returns
    -------
    alpha: array-like, float
        solar azimuth in degrees (zero in the south, increasing to 360 clockwise)
    """
    alpha = np.rad2deg(np.arccos(
      (np.sin(np.deg2rad(elevangle)) * np.sin(np.deg2rad(lat)) - np.sin(np.deg2rad(declination))) / 
        (np.cos(np.deg2rad(elevangle)) * np.cos(np.deg2rad(lat)))
    )) 
    alpha = (alpha * np.sign(hourangle) + 360) % 360

    return alpha

## Note that the function name starts with a `_`.
#  This indicates that the function is only intended to be used *internally*, so from 
#  other functions of this module, but not from users of the module.
#  This is only a convention, though, and highlights the intent of the programmer
#  to the user. Nevertheless, users can still call the function like any other funtion
#  from the module, if they choose to do so.
def _normalize_angle_difference(a, b):
    """
    Compute and normalize the difference between two angles 'a' and 'b' to the range [0, 360) degrees,
    where 0 represent south, and angles increase clockwise.

    Parameters
    ----------
    a (float): The first angle in degrees.
    b (float): The second angle in degrees.

    Returns
    -------
    float: The normalized angle difference between 'a' and 'b' in degrees, within the range [0, 360),
    where 0 represent south, and angles increase clockwise.
    
    This function calculates the difference between two angles 'a' and 'b' and ensures that
    the result is within the range [0, 360) degrees. It handles cases where the difference
    may cross the boundary of 360 degrees by adding or subtracting 360 as needed.
    """
    delta = a - b
    delta_normalized = np.where(delta < 0, delta + 360, delta)
    delta_normalized = np.where(delta_normalized >= 360, delta_normalized - 360, delta_normalized)
    return delta_normalized


def comp_cos_incidenceangle(elevangle, solarazimuth, incline, pvazimuth=0):
    """
    Compute the cosine of the incidence angle of direct sunlight onto an inclined surface

    Parameters
    ----------
    elevangle: array-like, float
        elevation angle of the sun in degrees
    solarazimuth: array-like, float
        solar azimuth as given by function `comp_solarazimuth()`
    incline: float
        incline of the surface in degrees [0, 90]

    Returns
    -------
    cos_theta: array-like, float
        cosine of incidence angle
    """
    cos_theta = (np.cos(np.deg2rad(incline)) * np.sin(np.deg2rad(elevangle)) + 
                 np.sin(np.deg2rad(incline)) * np.cos(np.deg2rad(elevangle)) * np.cos(np.deg2rad(_normalize_angle_difference(solarazimuth, pvazimuth))))
    cos_theta[cos_theta < 0] = 0
    return cos_theta

def comp_irradiance_incline(irradiance0, cos_incidenceangle, elevangle):
    """
    Compute direct sun irradiance onto an inclined surface

    Parameters
    ----------
    irradiance0: array-like, float
        irradiance onto horizontal surface
    cos_incidenceangle: array-like, float
        cosine of incidence angle onto surface
    elevangle: array-like, float
        elevation angle of the sun in degrees

    Returns
    -------
    irradiance_incline: array-like, float
        irradiance onto inclined surfaced
    """
    irradiance_incline = irradiance0 * cos_incidenceangle / np.sin(np.deg2rad(elevangle))
    return irradiance_incline

def parameterize_iswr_clearsky(elevangle, day_of_year, lat, declination):
    """
    Parameterize clear-sky solar irradiance on a horizontal surface. This is the maximum direct radiation to 
    expect on a given day, removing any other processes such as absorption and scatter by the atmosphere, etc.

    Parameters
    ----------
    elevangle: array-like, float
        solar elevation angle in degrees
    day_of_year: array-like, float
        day of the year (1--365)
    lat: array-like, float
        latitude in degrees
    declination: array-like, float
        solar declination in degrees
    
    Returns
    -------
    iswr_clearsky: array-like, float
        Clear-sky solar irradiance on a horizontal surface in W/m².
    """
    iswr_0 = 1361  # Solar constant in W/m²

    iswr_clearsky = iswr_0 * (1 + 0.033 * np.cos(2 * np.pi * day_of_year / 365)) * np.cos(np.deg2rad(lat)) * \
              np.cos(np.deg2rad(declination)) * np.sin(np.deg2rad(elevangle))
    iswr_clearsky[iswr_clearsky < 0] = 0

    return iswr_clearsky